# Commands -- parses topology and keyboard commands
#
# The topology file and the keyboard input file have the same format;
# all the same commands can be entered from either one.  Both
# are line-oriented (one command per line).
# Nodes (e.g., a, b) are referred to by their FishnetAddress (0..254).
#
#	// <comment>  -- any line starting with // is ignored
#	edge: a b [lossRate: <float>] [delay: <int>] [bw: <int>]
#		-- this creates an edge between a and b, with the
#		specified loss rate, delay (in milliseconds), and bw (in KB/s),
#		or changes the specifics for an existing link
#		defaults: 0 lossRate, 1 msec delay, and 10Kb/s bw
#	time: [+]x  -- any subsequent command is delayed until simulation/real
#			has reached x (or now + x, if + is used), in milliseconds from start
#	fail: a [b] -- this removes node a (if b is not specified) or an edge (if it is)
#	restart: a [b]  -- this restarts a node or edge.  previous information about
#		the node/edge is preserved
#	echo: text -- print the text 
#	exit  -- cleanly stop the simulation/emulation/ipaq run and print statistics
#	a: <msg>  -- deliver text <msg> to node a (for simulation mode only)
#	<msg> -- deliver text <msg> to this node (for emulation/ipaq mode only)
#		Note that msg cannot start with any keyword defined above
#
# To avoid a race condition with respect to starting up the user protocol code,
# the simulator will only process keyboard commands at time >  0.

# Implementation notes:
#	This file leverages the Ruby regular expression facility.  Please see
#	the Ruby book for more detailed information.  In general, using a regexp
#	parser will result in clearer code (and code that is easier to maintain)
#	than if we built a custom parser on our own.
#
#	We use inheritance to specialize the command parser for each mode
#	(simulation, emulation/ipaq, or trawler).  This is done by redefining
#	Commands::processLine, which invokes different sets of specific
#	parsers depending on the context.
#
#	All commands are processed in the present; when we see a "time" command,
#	we kick that back to the fishnet or trawler, to reinvoke when the time is right
#	(e.g., by using event processing).

module Fishnet

require "utility.rb"

# Commands -- routines for the command line parser
# Note that there's no local state, so no need for initialize
class Commands

	# parse a line for a time-related command
	# NOTE: internally we represent time as microseconds, so we have to
	# convert the parsed time, which is in milliseconds
	def parseTime(line, now)
		return nil unless line =~ /^time:\s*(\+)?(\d+)$/
		time = $2.to_i * 1000
		time += now if $1 == '+'
		return time
	end
	
	# parseCmd -- parse a line for a node command
	# 	first, ignore anything starting with // or #
	# 	then check for echo
	# 	then check if the user wants us to exit
	#	then, check if this is a simulation and we have num: msg
	#	otherwise, the line is the message to the node
	def parseCmd(line, indirect)
		line = line.strip
		return nil if line =~ %r{^//} # %r{} is for regexps with slashes. :)
		return nil if line =~ /^#/
		if line =~ /^echo:\s*(.*)$/
		  puts $1
		  return nil
		end
		$fishnet.stop if line =~ /^exit/

		return line unless indirect
		return [$1.to_i, $2] if line =~ /^(\d+):(.*)/

		return nil
	end
	
	# parse edge annotations, and return as a list
	def parseEdgeOptions(opts)
		lossRate = 0.0	# by default, lossless link
		delay = 1	# by default, 1 millisecond delay
		bw = 10		# by default, 10KB/s link

		lossRate = $1.to_f if opts =~ /lossRate:\s*([0-9\.]+(:?[eE]\d*)?)\s*(.*)$/
		opts = $3 if opts =~ /lossRate:\s*([0-9\.]+(:?[eE]\d*)?)\s*(.*)$/
		delay = $1.to_i if opts =~ /delay:\s*(\d+)\s*(.*)$/
		opts = $2 if opts =~ /delay:\s*(\d+)\s*(.*)$/
		bw = $1.to_i if opts =~ /bw:\s*(\d+)/

		lossRate = 0.0 if lossRate > 1
		bw = 1 if bw == 0

		return [lossRate, delay, bw]
	end
	
	# Parse a line of input for topology-related commands.
	# For each command, ask $topology to make the change to itself.
	# Note that topology in turn notifies $fishnet or $trawler about any changes.
	# Return true if we matched anything, since that means this isn't a keyboard
	#	message to be delivered to the node.
	def parseTopo(line)
		if line =~ /\A\s*edge:\s*(\d+)\s*(\d+)\s*(.*)$/
			a, b = $1.to_i, $2.to_i
			opts = $3
			if Fishnet.address?(a) and Fishnet.address?(b)
				options = parseEdgeOptions(opts)
				$topology.newEdge(a, b, options)
			else
				$stderr.print "No edge: ", line
			end
		elsif line =~ /\A\s*(fail|restart):\s*(\d+)\s*/
			node = $2.to_i
			if !Fishnet.address?(node) then
				$stderr.print "No node to fail: ", line
			elsif ($1 == "fail")
				$topology.failNode(node)
			else
				$topology.restartNode(node)
			end
		elsif line =~ /\A\s*(fail|restart):\s*(\d+)\s*(\d+)\s*/
			a, b = $2.to_i, $3.to_i
			if !Fishnet.address?(a) or !Fishnet.address?(b) then
				$stderr.print "No edge to fail: ", line
			elsif ($1 == "fail")
				$topology.failEdge(a,b)
			else
				$topology.restartEdge(a,b)
			end
		else
			return false
		end
		return true
	end
	
	# Process a single line of a topology file or keyboard input.
	# This implementation is the default, for emulation and IPAQ mode.
	#  For these, we consider just time and command options.
	# NOTE: we redefine this function for the simulator and the trawler, see below.
	def processLine(line, now)
 		$debug.print Debug::INPUT, 
 		             "Entering processLine with ", line, "\n"
		defer = parseTime(line, now)
		if !defer and (msg = parseCmd(line, false))
			$fishnet.node.onCommand(msg)
		end
		return defer
	end
	
	# Parse the remainder of a command file
	# if we encounter a "time" command, stop parsing and return when
	#	we should resume.
	def parseRemainder(now)
		@file.each { |x|
			defer = processLine(x, now)
			return defer if defer
		}
		return nil
	end
	
	# Process a topology file, by opening the file, and then using
	# parseRemainder on the whole file.
	def parseFile(fileName, now)
		if fileName then
			@file = File.new(fileName)
			return parseRemainder(now)
		else
			return nil
		end
	end

end  # Commands

# Specialized routines for parsing the topology file for the trawler
class TrawlerCmds < Commands

	# this registers that the trawler is interested in every topology change
	def initialize(allToAll)
		$topology = TrawlerTopology.new(allToAll)
	end
	
	# The trawler is only interested in time and topology commands, not
	# node messages.  You must type node messages to the individual nodes.
	def processLine(line, now)
		defer = parseTime(line, now)
		if !defer and !parseTopo(line) then
			$stderr.print "Unrecognized command ", line
		end
		return defer
	end
end

# Specialized routines for parsing the topology and keyboard commands for the
# simulator.
class SimulationCmds < Commands

	# this registers that the simulator is interested in certain topology changes
	def initialize
		$topology = SimTopology.new
	end
	
	# The simulator is interested in all types of commands
	def processLine(line, now)
		defer = parseTime(line, now)
		if !defer and !parseTopo(line) then
			if now == 0	# special case, to avoid a race with Node::start
				$stderr.print "Node message at time=0, ignored"
				return defer
			end
			node, msg = parseCmd(line, true)
			if !node
				# Not a command; do nothing
			elsif !Fishnet.address?(node)
				$stderr.print "No such node: ", line
			elsif node >= $fishnet.nodes.length then
				$stderr.print "Unrecognized command: ", line
			else
				$debug.puts Debug::ALL, "> #{node}: #{msg}"
				$fishnet.nodes[node].onCommand(msg)
			end
		end
		return defer
	end

end 	# SimulationCmds

end # module Fishnet

# vim: set ts=2 sw=2 noet :
