#!/usr/bin/ruby

require 'simulator'
require 'physical'

require 'getoptlong'

# add the project directory to the search path so we can load the node class.
$LOAD_PATH.push('../proj')
$LOAD_PATH.push('proj')
require 'node'


# usage:  print usage and exit.
def usage
	$stderr.print <<EOT
Usage: #{$0} subcommand [options]

subcommand must be one of:
  simulate NUM_NODES TOPOFILE  -- run simulation with NUM_NODES nodes locally 
	                                using topology file TOPOFILE
  emulate LOCALPORT HOST:PORT  -- register LOCALPORT at trawler at HOST:PORT
                                  to run with other nodes
                                  (example: HOST is 'attu1', not 'localhost')
  ipaq IPAQ_NUM:LOCALPORT      -- broadcast to other nodes via LOCALPORT where
                                  the number of your ipaq is IPAQ_NUM

Options:
  --timescale=SCALE    Multiply time by SCALE (can be fractional)
                       [Simulator only]
  --input=FILE         Read commands from FILE (on Unix, this is equivalent to
                       appending " < file" to the command line.)

Examples:
  #{$0} simulate three.topo --timescale=0.5
  #{$0} emulate 8001 localhost:8000
EOT
	exit 1
end

# die:  print error message, help pointer, and exit.
def die(s)
	$stderr.puts "Fatal: " + s
	$stderr.puts "Use '#{$0} --help' for help."
	exit 1
end

# parseport:  verify a string is a valid port and return it.
def parseport(port)
	port = port.to_i
	die("bad port '#{port}'.") unless port > 0
	return port
end

# parseaddr:  verify a string is a valid host:port and return them.
def parseaddr(hostport)
	address, port = hostport.split(/:/)
	unless port
		port = address
		address = 'localhost'
	end
	die("malformed address '#{hostport}'.") unless address and port
	port = parseport(port)
	return address, port
end

def main
	timescale = nil

	$stdout.sync = true

	opts = GetoptLong.new(
		[ "--input",     "-i", GetoptLong::REQUIRED_ARGUMENT ],
		[ "--timescale", "-s", GetoptLong::REQUIRED_ARGUMENT ],
		[ "--help",      "-h", GetoptLong::NO_ARGUMENT       ]
	)

	begin
		opts.each { |opt, arg|
			case opt
			when '--input'
				$fishInput = File.open(arg, "r")
			when '--timescale'
				timescale = arg
			when '--help'
				usage; exit 1
			end
		}
	rescue
		die("error parsing options.")
	end

	subcommand = ARGV.shift
	die("must specify subcommand.") unless subcommand

	$debug = Debug.new
	case subcommand
	when /^sim/  # simulator mode
		num_nodes = ARGV.shift.to_i
		die("must specify number of nodes for simulation.") unless (num_nodes > 0)
		topofile = ARGV.shift
		die("must specify topology file for simulation.") unless topofile
		fishnet = Fishnet::Simulator.new(num_nodes - 1, topofile)
		fishnet.timescale = timescale.to_f if timescale
	when /^em/  # emulator mode
		port = ARGV.shift
		die("must specify port.") unless port
		lPort = port.to_i
		trawler = ARGV.shift
		die("must specify trawler address.") unless trawler
		tAddress, tPort = parseaddr(trawler)

		fishnet = Fishnet::Emulator.new(tAddress, tPort, lPort)
	when /^i/   # ipaq mode (allow just "i" 'cause it's painful to type on ipaqs)
		#		lPort = Fishnet::Ipaq::DefaultPort
		address = ARGV.shift
		unless address
			die("must specify local address.")
		end
		lAddress, lPort = parseaddr(address)
		lAddress = lAddress.to_i
		fishnet = Fishnet::Ipaq.new(lAddress, lPort)
	else
		$stderr.puts "Unknown subcommand #{subcommand}."
		usage
	end

	$fishnet = fishnet  # some files still depend on this global.

	fishnet.start		# start the simulator/emulator/ipaq execution
end

main

# vim: set ts=2 sw=2 noet :
