# manager -- defines generic routines for running network code under fishnet.

# Manager is an abstract class representing various ways (simulation,
# emulation, ipaq, planetlab) to execute network code.
#
# All subclasses implement:
#   start() -- called by main when the Fishnet is ready to begin
#   sendPkt(to, pkt)  -- called by node software to send a packet
#                     -- to is a FishNet Address, pkt is a binary string,
#                        defined in packet.rb
#   wakeupIn(x) -- called by node software to set a wakeup call,
#                  set a timer to go off in x milliseconds, return immediately
#   pong -- make a "pong" sound on the IPAQ, and print "PONG" in
#           simulation and emulation mode
#   now -- return the current time in milliseconds
#          (real time on the IPAQ, simulated time on the simulator)
#
# All subclasses make upcalls to class Node; Node is implemented as a state
# machine, so all of these routines return after doing their immediate work:
#   start -- when Fishnet is ready for use
#   onReceive(from, pkt)  -- when a packet arrives, notify Node
#   onKeyboard(msg) -- when a line of text (a command) is ready from the input
#   onTimer(now) -- when a wakeup alarm triggers, now is current time (in msec)
#

require 'packet'
require 'utility'
require 'topology'
require 'commands'

module Fishnet
class Manager
	def initialize
		@numSent = 0
		@start = now
	end
	
	def packet?(x)		# check the packet size
		(1...MaxPacketSize) === x.length
	end
	
	# node puts.  specialized in the simulator.
	def node_log(debug_level, from, *args)
		args.push "\n"
		$debug.print debug_level, *args
	end

	# in the generic class, just check the arguments
	def sendPkt(to, pkt)
		raise ArgumentError if !Fishnet.address?(to) or !packet?(pkt)
		@numSent += 1
	end
	
	# called by node to register a wakeup alarm at time t
	# overridden by subclasses.
	def addTimerAt(node, t, callback)
		raise ArgumentError if t < @now
	end

	# called by node to register a wakeup alarm x milliseconds in the future
	def addTimer(node, dt, callback)
		addTimerAt(node, @now + dt*1000, callback)
	end
		
	# on the IPAQ, we redefine this to play a sound file
	def pong
		"PONG!!"
	end
	
	def stop
		$stderr.print "Fishnet exiting at time: ", 
			self.now() - @start, " msec.", 
			"\nNumber of packets sent: ", @numSent, "\n"
		exit(1)
	end

	def log(debug_level, *args)
		$debug.print debug_level, "fishnet: "
		$debug.print debug_level, *args
		$debug.print debug_level, "\n"
	end
end

class NodeBase
	attr_reader :fishnet, :addr

	def initialize(fishnet, addr)
		@fishnet = fishnet
		@addr = addr
	end

	# proxy(foo) makes a new method foo(*args)
	# that calls fishnet.foo(@addr, *args).
	def NodeBase.proxy(*funcs)
		funcs.each { |f|
			class_eval "def #{f}(*args); @fishnet.send(:#{f}, @addr, *args); end"
		}
	end
	proxy :addTimer, :addTimerAt, :sendPkt

	def log(debug_level, *args)
		@fishnet.node_log(debug_level, @addr, *args)
	end
end

end # module Fishnet

# vim: set ts=2 sw=2 noet :
