# Packet -- defines the Fishnet packet header, used for
# 	communicating between Fishnet nodes.

module Fishnet

# Some global constants
BroadcastAddress = 255    # 255 signifies a broadcast
MaxAddress = 255
MaxPacketSize = 128       # bytes, kept small on purpose
MaxTTL = 15               # max hop count

# test if a Fishnet address is legal, from 0..255
def Fishnet.address?(x)
	(0..MaxAddress) === x
end

# Protocols we currently support
# Note that we'll add to these as the quarter progresses,
# so do not modify this file!
PingPkt      = 0   # packet contents is a text message
PingReplyPkt = 1
LinkInfoPkt  = 2   # packet contents contain the list of neighbors for the source address
NamePkt      = 3   # packet contents contains the machine name for the source address
TransportPkt = 4   # packet contents are for reliable transport

# test if a protocol is legal
def Fishnet.protocol?(x)
	(PingPkt..TransportPkt) === x
end

# for printing error messages
def Fishnet.protocolToS(x)
	case x
	when PingPkt
		"Ping"
	when PingReplyPkt
		"Ping reply"
	when LinkInfoPkt
		"Link state packet"
	when NamePkt
		"Name packet"
	when TransportPkt
		"Transport"
	else
		"Unknown protocol"
	end
end

# test if this is a legal TTL value
def Fishnet.ttl?(x)
	(1..MaxTTL) === x
end

# test if this packet is legal
def Fishnet.packet?(to, from, protocol, ttl, length)
	return (address?(to) and address?(from)
	       protocol?(protocol) and ttl?(ttl) and
				 length <= MaxPacketSize)
end

# Packet manages the format of packets on the wire
#	destination address: 8 bits
#	source address: 8 bits
#	ttl (time to live) (8 bits)
#	protocol: 8 bits
#	packet length: 16 bits
#	packet id: 32 bits
#	payload: <up to (MaxPacketSize-(the above)) bytes>

class Packet
	@@Format = "CCCCnN"   # packet header in binary
	@@HeaderSize = 10     # number of bytes added by header
	
	# create a packet to be sent
	def initialize(to, from, protocol, ttl, seq, msg)
		raise ArgumentError if !Fishnet.packet?(to, from, 
				                                    protocol, ttl, msg.length + @@HeaderSize)
		@dest = to
		@src = from
		@protocol = protocol
		@msg = msg
		@ttl = ttl
		@seq = seq
	end
	
	# allow various routines to directly access the contents of the packet
	attr_reader :dest, :src, :protocol, :ttl, :seq, :msg
	attr_writer :ttl, :dest
	
	# print the packet
	def to_s
		"Packet: " + 
			(@src.to_s + "->" + @dest.to_s) +
			" protocol: " + Fishnet::protocolToS(@protocol) + 
			" ttl: " + @ttl.to_s +
			" seq: " + @seq.to_s +
			" contents: " + @msg.inspect
	end
	
	# convert the packet into a binary string for sending over the wire
	def pack
		[@dest, @src, @ttl, @protocol, 
			@msg.length + @@HeaderSize, 
			@seq].pack(@@Format) + @msg
	end
	
	# This is a class method to unpack a binary string into a packet.
	# This assumes the binary string begins with the packet header
	# Returns the new packet, or nil if the binary string is obviously corrupted.
	def Packet.unpack(str)
		dest, src, ttl, protocol, length, id = str.unpack(@@Format)
		msg = str[@@HeaderSize...str.length]
		return nil unless Fishnet.packet?(dest, src, protocol, ttl, str.length)
		return Packet.new(dest, src, protocol, ttl, id, msg)
	end

end

#This manages a link state packet, to convey a list of neighbors around the
# network for building routing tables.
#
# The format of the link state packet is very simple:
#	a variable length array of <fishnet address>
#
#  Note that this header is assumed to be the "msg" contents of a class Packet,
#  so it does not need source, destination, TTL, sequence #, etc., information.

class LinkState
	@@MaxNeighbors = 50		# so we don't exceed the packet size
	
	def initialize(neighbors)
		neighbors.each {|x| throw ArgumentError if !Fishnet.address?(x)}
		raise ArgumentError if neighbors.length > @@MaxNeighbors
		@neighbors = neighbors
	end
	
	attr_reader :neighbors
	
	def pack
		if @neighbors.length > 0
			@neighbors.pack("C"*@neighbors.length)
		else
			""		# return the empty string, signifying no neighbors
		end
	end
	
	def LinkState.unpack(str)
		if str.length == 0
			neighbors = []
		else
			neighbors = str.unpack("C"*str.length)
			# if any neighbor is invalid, assume the packet is corrupted and discard
			neighbors.each { |x| return nil if !Fishnet.address?(x)}
		end
		return LinkState.new(neighbors)
	end	
end

# This conveys the header for reliable message transfer.
# This is carried in the payload of a Packet, and in turn the data being
# transferred is carried in the payload of the Transport packet.
#types of packets

class Transport
	@@Format = "nnnnN"
	@@Size = 12

	MaxPayload = MaxPacketSize - (@@Size + 12)
	MaxPortNum = 255	# port numbers range from 0..255

	SYN  = 0   # connection setup
	ACK  = 1   # ack (note that the ack # comes back in the seqNo)
	FIN  = 2   # close the connection
	DATA = 3   # regular data
	
	def initialize(srcPort, destPort, type, window, seqNo, payload)
		@srcPort = srcPort    # port to use for replies back to source
		@destPort = destPort  # pkt is delivered to this port
		@type = type          # syn, ack, fin, data
		@window = window      # advertised window (for flow control)
		@seqNo = seqNo        # sequence number in bytes
		                      # (for acks, the last byte received)
		@payload = payload    # data being transferred, if any
		raise ArgumentError if ((payload.length > 0 and type != DATA) or
						payload.length > MaxPayload)
	end
	
	attr_reader :srcPort, :destPort, :type, :seqNo, :window, :payload
	
	def pack
		[@srcPort, @destPort, @type, @window, @seqNo].pack(@@Format) + @payload
	end
	
	def Transport.unpack(str)
		return nil unless str.length >= @@Size
	        srcPort, destPort, type, window, seqNo = str.unpack(@@Format)
		payload = str[@@Size...str.length]
		return nil if srcPort > MaxPortNum or destPort > MaxPortNum
		if type == SYN or type == ACK or type == FIN
			return nil if payload.length > 0
		elsif type == DATA
			return nil if payload.length > MaxPayload
		else
			return nil		# illegal type
		end
		return Transport.new(srcPort, destPort, type, window, seqNo, payload)
	end
end

end # module Fishnet

# vim: set ts=2 sw=2 noet :
