# Topology -- routines to keep track of a dynamic, annotated topology
#	Used in the simulator and the trawler.

module Fishnet
# Edge stores the specifics about each edge in the topology.  Edges can
# be temporarily disabled (set @live to false), and they record when the
# next packet can be sent along the edge.
class Edge
	def initialize(a, b)
		@a = a		# fishAddresses of the two vertices of the edge
		@b = b
			# when can we send on this link?  Kept separately in each direction	
		@nextPkt = { a => 0, b => 0}	
		@live = true				# if the edge has failed
	end
	
	# schedulePkt -- Figure out when (in microseconds!) a packet will arrive 
	# at the destination, given a link's bw/propagation delay characteristics.  
	# For now, only used in simulation mode. 
	def schedulePkt (src, size, now)
		raise ArgumentError if size > MaxPacketSize
		raise ArgumentError if !@nextPkt.has_key?(src)
		
		result = [now, @nextPkt[src]].max
		@nextPkt[src] = result + size*1000/@bw
		if (!@live or rand() < @lossRate) then
			return nil			# packet was dropped
		else
			return @nextPkt[src] + @delay*1000
		end
	end
	
	# we make these public, so that Topology can edit them
	attr_reader :a, :b
	attr_accessor :lossRate, :delay, :bw, :live
end	# Edge

# Topology -- keep track of the connections between nodes
# 	for the simulated and emulated environments
class Topology
	# initialize the topology; initially empty
	# if allToAll is set, connect everything to everything else by default
	def initialize(allToAll)
		@edges = []			# list of edges
		@failNodes = []			# list of failed nodes
		@allToAll = allToAll		# connect all nodes by default
	end
	
	def to_s
		if @allToAll
			str = "all to all"
		else
			strings = @edges.collect { |x|
				if liveNode?(x.a) and liveNode?(x.b)
					x.a.to_s + "->" + x.b.to_s
				else
					nil
				end
			}
			str = strings.compact.join(", ")
		end
		'[' + str + ']'
	end
	
	# is (a, b) in the topology?  if so, return the edge
	def edge?(a, b)
		return Edge.new(a, b) if @allToAll
		@edges.each {|e|
			return e if (e.a == a && e.b == b) or 
					(e.a == b && e.b == a)
			}
		return nil	# not found
	end
	
	# is (a, b) in the topology, and alive?
	def liveEdge?(a, b)
		edge = edge?(a,b)
		return nil if !edge or !edge.live or 
				!liveNode?(a) or !liveNode?(b)
		return edge
	end
	
	# is this node still alive, or has it failed?
	def liveNode?(a)
		!@failNodes.include?(a)
	end
				
	# create or update an Edge, with the specified options
	def newEdge(a, b, options)
		if (edge = edge?(a, b)) then
			edge.live = true
		else
			edge = Edge.new(a, b)
			@edges.push edge
		end
		edge.lossRate, edge.delay, edge.bw = options
	end
	
	def chgEdge(a, b, live)
		if (edge = edge?(a, b)) then
			edge.live = live
		else
			$stderr.print "No edge to change ", a, " ", b
		end
	end
	
	# mark the edge as failed, so we don't use it
	def failEdge(a, b)
		chgEdge(a, b, false)
	end
	
	# edge has recovered
	def restartEdge(a, b)
		chgEdge(a, b, true)
	end
	
	# mark the node as failed, so we don't use any edge that goes through it
	def failNode(a)
		@failNodes.push(a)
	end
	
	# mark it as ok, so we can use it again
	def restartNode(a)
		@failNodes.delete(a)
	end
end

class SimTopology < Topology
	def initialize
		super(false)
	end
	
	def restartNode(a)
		super(a)
		$fishnet.restartNode(a)	# tell the fishnet to restart the node
	end
end

# Use these hooks to alter the neighbor list on each of the emulated nodes
class TrawlerTopology < Topology

	def newEdge(a, b, options)
		super(a,b,options)
		$trawler.startEdge(a,b)
	end
	
	def chgEdge(a, b, live)
		super(a,b, live)
		live ? $trawler.startEdge(a,b) : $trawler.failEdge(a,b)
	end

	# note: delete from the topology after we notify the trawler
	# so we know which edges to delete	
	def failNode(a)
		$trawler.failNode(a)
		super(a)			
	end
	
	def restartNode(a)
		super(a)
		$trawler.restartNode(a)
	end
end
end # module Fishnet

# vim: set ts=2 sw=2 noet :
